#
# BlueStole - BlueCloth Wrapper
#
# (C)2005 H.Miyamoto (aka "Yuukis") <Ys@PixyGarden.net>
# You can redistribute it and/or modify it under GNU GPL2.
#
# $Id: bluestole.rb,v 1.5 2006/01/07 13:57:15 ys Exp $
#

require 'bluecloth'

class BlueClothPlus < BlueCloth
	def initialize(str)
		@include_htmls = {}
		include_id = 0

		x = str.gsub(/^<html>\n*(.*?)\n*<\/html>/m) {
			include_id += 1
			@include_htmls[include_id]= $1
			"# #{Process.pid}_#{include_id}"
		}
			
		super(x)
	end

	def to_html
		x = super()
		x.gsub(/<h1>#{Process.pid}_(\d+)<\/h1>/) {
			@include_htmls[$1.to_i]
		}
	end
end

class BlueStole

	VERSION = '0.6'
	raise 'Use Ruby 1.8.4 or later' if "1.8.4" <= RUBY_VERSION

	attr_accessor :deflist, :table, :fn, :hxid,
		:sub_us, :sub_ees, :image_info

	def initialize(str)
		@str = str
		@pid_hex = Digest::MD5.hexdigest("%%BlueStole#{Process.pid}%%")
		setup
	end

	def to_html
		ret = @str.dup

		# Pre functions
		ret = escape_footnote(ret) if @fn
		ret = sub_underscore(ret) if @sub_us
		ret = sub_deflist(ret) if @deflist
		ret = sub_table(ret) if @table

		# Markdown convert
		ret = translate(ret)

		# Post functions
		ret = sub_empty_element_suffix(ret) if @sub_ees
		ret = titlize(ret) if @hxid
		ret = add_image_info(ret) if @image_info
		ret = make_footnote(ret) if @fn

		ret
	end


	private

	def setup
		@deflist, @fn, @hxid = true, true, true
		@sub_us, @sub_ees, @image_info = true, true, true

		begin
			require 'rt/rtparser'
			require 'rt/rt2html-lib'

			@table = true
		rescue LoadError
			@table = false
		end
	end

	def escape_footnote(str)
		str.gsub(/\(\*/, @pid_hex)
	end

	def make_footnote(str)
		ret = str.gsub(/#{@pid_hex}/, '(*')
		fn_id = '00'
		footnote = ''

		ret = ret.gsub(/\(\*(.*?)\)/) {
			fn_id.succ!
			fn_string = $1

			footnote << %Q( <li><p class="fn"><a id="fn_#{fn_id}_02")
			footnote << %Q( name="fn_#{fn_id}_02")
			footnote << %Q( href="#fn_#{fn_id}_01">)
			footnote << %Q( #{fn_id.to_i.to_s}</a> #{fn_string}</p></li>\n)
			# id.to_i.to_s for '02' => '2'

			r = ''
			r << %Q(<sup><a id="fn_#{fn_id}_01")
			r << %Q( name="fn_#{fn_id}_01")
			r << %Q( href="#fn_#{fn_id}_02")
			r << %Q( title="#{fn_string}">)
			r << %Q(*#{fn_id.to_i.to_s}</a></sup>)
			# id.to_i.to_s for '02' => '2'
		}

		x = ''
		if footnote != ""
			x = %Q(\n\n<ol class="fn">\n#{footnote}</ol>\n)
		end

		ret + x
	end

	def titlize(str)
		article_id = '00'
		r = ''

		str.each {|line|
			if line !~ /<\/a>/ && line =~ /^(.*?<h\d>)(.*?)(<\/h\d>.*?)/
				article_id.succ!

				a_ids = %Q(<a id="id_#{article_id}" name="id_#{article_id}">)
				a_ide = %Q(</a>)

				line = $1 + a_ids + $2 + a_ide + $3
			elsif line =~ /^(.*?<h\d>.*?)<a href="(.*?)">(.*?)<\/a>(.*?<\/h\d>.*?)/
				article_id.succ!

				a_ids = %Q(<a id="id_#{article_id}" name="id_#{article_id}" href="#{$2}">)
				a_ide = %Q(</a>)

				line = $1 + a_ids + $3 + a_ide + $4
			end
			r << line
		}

		r
	end

	def sub_underscore(str)
		str.gsub(/\\*_/, '\_')
	end

	def sub_empty_element_suffix(str)
		str.gsub(/\/>/, ' />')
	end

	def sub_deflist(str)
		deflist_regex = %r{((.*:\n(\s*.*\n)*?\n\n*)+)}
		str.gsub(/#{deflist_regex}/) { _sub_deflist($1) }
	end

	def _sub_deflist(str)
		ret = ''
		_deflist_regex = %r{(.*):\n((\s*.*\n)*?)\n\n*}
		ret << str.gsub(_deflist_regex) {
			dt = translate($1.dup)
			dd = translate($2.dup.gsub(/^\s*/, '').chomp)
			x = '<dt>' + dt.gsub(/<p>(.*?)<\/p>$/, '\1') + '</dt>' + "\n"
			x << '<dd>' + dd.gsub(/<p>(.*?)<\/p>$/m, '\1') + '</dd>' + "\n"
			x
		}
		"<html>\n<dl>\n" + ret + "</dl>\n</html>\n"
	end

	def sub_table(str)
		table_regex = %r{((^\|.*\n)+)}
		str.gsub(/#{table_regex}/) { _sub_table($1) }
	end

	def _sub_table(str)
		visitor = RT::RT2HTMLVisitor.new
		"<html>\n" + visitor.visit(RT::RTParser.parse(str.gsub(/^\|/, ''))) + "</html>\n"
	end

	def translate(str)
		bc = BlueClothPlus.new(str)
		bc.to_html
	end

	def add_image_info(str)
		str.gsub(/(<img src="(.*?)" alt="(.*?)" \/>)/) {
			all, src, alt = $1, $2, $3
			if File.exist?(src)
				height, width = image_info(File.open(src))
				if height
					%Q(<img src="#{src}" alt="#{alt}" height="#{height}" width="#{width}" />)
				else
					all
				end
			else
				all
			end
		}
	end

	# code from image.rb(Revesion: 1.22) tdiary's plugin. thanks.
	def image_info(f)
		image_height = nil
		image_width = nil
	
		sig = f.read( 24 )
		if /\A\x89PNG\x0D\x0A\x1A\x0A(....)IHDR(........)/on =~ sig
			image_width, image_height = $2.unpack( 'NN' )
	
		elsif /\AGIF8[79]a(....)/on =~ sig
			image_width, image_height = $1.unpack( 'vv' )
	
		elsif /\A\xFF\xD8/on =~ sig
			data = $'
			until data.empty?
				break if data[0] != 0xFF
				break if data[1] == 0xD9
	
				data_size = data[2,2].unpack( 'n' ).first + 2
				if data[1] == 0xC0
					image_height, image_width = data[5,4].unpack('nn')
					break
				else
					if data.size < data_size
						f.seek(data_size - data.size, IO::SEEK_CUR)
						data = ''
					else
						data = data[data_size .. -1]
					end
					data << f.read( 128 ) if data.size < 4
				end
			end
		end
	
		return image_height, image_width
	end
end

